#include "stdafx.h"
#include "InstallController.h"
#include "InstallPrepareModule.h"
#include "UpgradePrepareModule.h"
#include "RemovePrepareModule.h"
#include "ScheduleModuleWnd.h"
#include "AgentDLL.h"
#include <atlchecked.h>
#include <memory>
#include "util.h"
#include "IPC/IPCCopyData.h"
#include "Packet.h"
#include <AtlPath.h>
#include "LauncherManager.h"

using namespace tipbase;

const DWORD POSTMESSAGE_INTERVAL = 250;
class MyInstallCallback : public IInstallModuleCallback
{
public:
	MyInstallCallback( ScheduleModuleWnd *scheduleWnd, const TCHAR *guid ) : m_scheduleWnd(scheduleWnd), m_guid(guid)
	{
	}

	virtual CALLBACK_RETURN OnStart()
	{
		if( FALSE == MyPostMessage( m_scheduleWnd->m_hWnd, WM_INSTALL_START, 0, (LPARAM)m_guid.c_str()  ) )
			return CALLBACK_RETURN_ERROR;

		return CALLBACK_RETURN_CONTINUE;
	}

	virtual CALLBACK_RETURN OnFinished(CALLBACK_FINISHED_STATUS status, ERRORINFO *errorInfo )
	{
		CALLBACK_FINISHED_INFO *callbackInfo = new CALLBACK_FINISHED_INFO;
		if( NULL == callbackInfo )
			return CALLBACK_RETURN_CONTINUE;

		callbackInfo->SetFinishedStatus( status );
		callbackInfo->SetGuid(m_guid.c_str());
		callbackInfo->SetErrorInfo(errorInfo);

		if( FALSE == MyPostMessage( m_scheduleWnd->m_hWnd, WM_INSTALL_FINISH, (WPARAM)0, (LPARAM)callbackInfo  ) )
			return CALLBACK_RETURN_ERROR;

		return CALLBACK_RETURN_CONTINUE;
	}
private:
	ScheduleModuleWnd*	m_scheduleWnd;
	const tstring		m_guid;
};

class MyPrepareCallback : public IPrepareModuleCallback
{
public:
	MyPrepareCallback( ScheduleModuleWnd *scheduleWnd, const TCHAR *guid) : m_scheduleWnd(scheduleWnd), m_guid(guid), m_lastTickCount(0)
	{
	}

	virtual CALLBACK_RETURN OnStart()
	{
		if( FALSE == MyPostMessage( m_scheduleWnd->m_hWnd, WM_PREPARE_START, (WPARAM)0, (LPARAM)m_guid.c_str()  ) )
			return CALLBACK_RETURN_ERROR;

		return CALLBACK_RETURN_CONTINUE;
	}

	void PostMessageWithInterval( UINT msg, CALLBACK_PROGRESS_STATUS status, DownloadSize *downloadSize )
	{
		DWORD currentTickCount = ::GetTickCount();
		if( (currentTickCount - m_lastTickCount) > POSTMESSAGE_INTERVAL )
		{
			MyPostMessage( m_scheduleWnd->m_hWnd, msg, (WPARAM)status, (LPARAM)downloadSize  );
			m_lastTickCount = currentTickCount;
		}
	}

	virtual CALLBACK_RETURN OnProgress(	CALLBACK_PROGRESS_SIZE currentSize,
		CALLBACK_PROGRESS_SIZE maxSize,
		CALLBACK_PROGRESS_STATUS status )
	{
		if( GetModule()->IsCancel() == TRUE )
		{
			return CALLBACK_RETURN_CANCEL;
		}

		DownloadSize *downloadSize = new DownloadSize;
		downloadSize->currentSize = currentSize;
		downloadSize->maxSize = maxSize;
		Checked::tcscpy_s(downloadSize->guid, GUID_SIZE, m_guid.c_str() );

		if( GetModule()->IsPending() )
			status = CALLBACK_PROGRESS_PENDING;

		PostMessageWithInterval( WM_PREPARE_RECEIVE, status, downloadSize );
		return CALLBACK_RETURN_CONTINUE;
	}

	virtual CALLBACK_RETURN OnFinished( CALLBACK_FINISHED_STATUS status, ERRORINFO *errorInfo )
	{
		CALLBACK_FINISHED_INFO *callbackInfo = new CALLBACK_FINISHED_INFO;
		if( NULL == callbackInfo )
			return CALLBACK_RETURN_CONTINUE;

		callbackInfo->SetFinishedStatus( status );
		callbackInfo->SetGuid(m_guid.c_str());
		callbackInfo->SetErrorInfo(errorInfo);

		if( FALSE == MyPostMessage( m_scheduleWnd->m_hWnd, WM_PREPARE_FINISH, (WPARAM)0, (LPARAM)callbackInfo ) )
			return CALLBACK_RETURN_ERROR;

		return CALLBACK_RETURN_CONTINUE;
	}

private:
	ScheduleModuleWnd*	m_scheduleWnd;
	DWORD				m_lastTickCount;
	const tstring		m_guid;
};

InstallController::InstallController()
{
	m_currentPrepareModule = NULL;
	m_currentInstallModule = NULL;
	m_installcontrollercallback = NULL;
	m_scheduleModuleWnd = NULL;
	m_prepareThreadPool = NULL;
	m_installThreadPool = NULL;

	DWORDLONG tickCount = GetTickCount();
	CString launcherName;
	launcherName.Format(_T("%08x"), tickCount );

	SetLauncherName(launcherName);
	m_bLauncherStart = StartLauncher();
}

InstallController::~InstallController()
{
	KillLauncher();

	m_currentPrepareModule = NULL;
	m_currentInstallModule = NULL;
	m_installcontrollercallback = NULL;
	m_scheduleModuleWnd = NULL;

	NPDeleteThreadPool(m_prepareThreadPool);
	NPDeleteThreadPool(m_installThreadPool);
	if( m_scheduleModuleWnd )
	{
		delete m_scheduleModuleWnd;
	}
}

BOOL InstallController::StartLauncher()
{
	return LauncherManager::GetInstance()->StartLauncher(GetLauncherName());
}

BOOL InstallController::KillLauncher()
{
	return LauncherManager::GetInstance()->KillLauncher();
}

BOOL InstallController::CheckLauncher()
{
	return LauncherManager::GetInstance()->CheckLauncher();
}

BOOL InstallController::Create()
{
	if( FALSE == m_bLauncherStart )
	{
		ATLTRACE( _T("Launcher Start Error\n") );
		return FALSE;
	}

	if( FALSE == CheckLauncher() )
	{
		ATLTRACE( _T("CheckLauncher Error\n") );
		return FALSE;
	}

	m_scheduleModuleWnd = new ScheduleModuleWnd(this);
	ATLASSERT( NULL != m_scheduleModuleWnd );
	if( NULL == m_scheduleModuleWnd->Create(HWND_MESSAGE) )
	{
		ATLTRACE( _T("m_scheduleModuleWnd Create Error\n") );
		return FALSE;
	}

	m_currentPrepareModule = NULL;
	m_currentInstallModule = NULL;

	m_prepareThreadPool = NPCreateThreadPool(1);
	m_installThreadPool = NPCreateThreadPool(1);

	ATLASSERT( NULL != m_prepareThreadPool || NULL != m_installThreadPool );
	if( NULL == m_prepareThreadPool || NULL == m_installThreadPool )
		return FALSE;

	return TRUE;
}

BOOL InstallController::Remove( InstallProgramInformationVector infoArray, 
							   IInstallControllerCallback *InstallControllerCallback )
{
	InstallProgramInformationVector::iterator i;
	for( i = infoArray.begin(); i != infoArray.end(); i++ )
	{
		ProgramInformation &info = *i;
		info.install_process_type = UninstallType;
	}

	return Install( infoArray, InstallControllerCallback );
}


BOOL InstallController::Install( InstallProgramInformationVector infoVector, 
								 IInstallControllerCallback *InstallControllerCallback)
{
	ATLASSERT( NULL != m_scheduleModuleWnd );
	if( NULL == m_scheduleModuleWnd )
	{
		ATLTRACE( _T("m_scheduleModuleWnd is Not Created\n") );
		return FALSE;
	}

	m_installcontrollercallback = InstallControllerCallback;
	ATLASSERT( NULL != m_prepareThreadPool || NULL != m_installThreadPool );
	if( NULL == m_prepareThreadPool || NULL == m_installThreadPool )
		return FALSE;

	if( infoVector.empty() == true || NULL == InstallControllerCallback )
	{
		return FALSE;
	}

	if( FALSE == MakePrepareModuleQueue( infoVector ) )
	{
		return FALSE;
	}

	return ProcessPrepare();
}

NAVERAPP_INSTALL_STATUS GetInstallStatusWithGuid(const TCHAR *guid)
{
	AgentDLLWrapper *agentDLL = AgentDLLManager::Find( guid );
	ATLASSERT( NULL != agentDLL );
	if( NULL == agentDLL )
	{
		return NotInstalled;
	}

	NAVERAPP_INSTALL_STATUS status;
	if( FALSE == agentDLL->GetInstalledStatus(status) )
	{
		return NotInstalled;
	}

	return status;
}

NAVERAPP_INSTALL_STATUS InstallController::GetInstalledStatus(const TCHAR *guid)
{
	return GetInstallStatusWithGuid(guid);
}

IPrepareModule *InstallController::CreateNewPrepareModule( ProgramInformation programInformation )
{
	IPrepareModule *prepareModule = NULL;

	switch(programInformation.install_process_type)
	{
	case InstallType:
		prepareModule = new InstallPrepareModule( programInformation.guid, programInformation.url, programInformation.path, programInformation.execute_option );
		break;
	case UpgradeType:
		prepareModule = new UpgradePrepareModule( programInformation.guid, programInformation.execute_option );
		break;
	case UninstallType:
		prepareModule = new RemovePrepareModule( programInformation.guid, programInformation.execute_option );
		break;
	default:
		break;
	}

	if( prepareModule )
		prepareModule->SetCallback( new MyPrepareCallback(this->m_scheduleModuleWnd, prepareModule->GetGuid()) );

	return prepareModule;
}

BOOL InstallController::MakePrepareModuleQueue(InstallProgramInformationVector infoVector)
{
	int maxSize = infoVector.size();

	for( int i = 0; i < maxSize; i++ )
	{
		IPrepareModule *prepareModule = CreateNewPrepareModule(infoVector[i]);
		if( NULL == prepareModule )
			continue;

		m_prepareModuleQueue.push_back(prepareModule);
	}

	return TRUE;
}

BOOL InstallController::NotifyPrepareSchedule()
{
	return MyPostMessage( m_scheduleModuleWnd->m_hWnd, WM_SCHEDULE_PREPAREMODULE, (WPARAM)0, (LPARAM)0 ); 
}

BOOL InstallController::NotifyInstallSchedule()
{
	return MyPostMessage( m_scheduleModuleWnd->m_hWnd, WM_SCHEDULE_INSTALLMODULE, (WPARAM)0, (LPARAM)0 ); 
}

BOOL InstallController::SetPending( const TCHAR *guid, BOOL flag )
{
	BOOL preparePending = PreparePending(guid,flag);
	BOOL installPending = InstallPending(guid,flag);

	NotifyPrepareSchedule();
	NotifyInstallSchedule();
	return ((preparePending == TRUE) || (installPending == TRUE));
}

BOOL InstallController::PreparePending( const TCHAR *guid, BOOL flag )
{
	IPrepareModule *prepareModule = NULL;
	if( m_currentPrepareModule && 
		_tcscmp( m_currentPrepareModule->GetGuid(), guid ) == 0 )
	{
		m_currentPrepareModule->SetPending(flag);
		return TRUE;
	}

	PrepareModuleQueue::iterator it = m_prepareModuleQueue.begin();
	BOOL ret = FALSE;
	for( ; it != m_prepareModuleQueue.end(); it++ )
	{
		IPrepareModule *module = (*it);
		if( _tcscmp( module->GetGuid(), guid ) != 0 )
		{
			continue;
		}

		module->SetPending(flag);
		ret = TRUE;
		break;
	}
	return ret;
}

BOOL InstallController::InstallPending( const TCHAR *guid, BOOL flag )
{
	IInstallModule *installModule = NULL;
	if( m_currentInstallModule && 
		_tcscmp( m_currentInstallModule->GetGuid(), guid ) == 0 &&
		m_currentInstallModule->GetModuleStatus() == ModuleStatusINIT )
	{
		m_currentInstallModule->SetPending(flag);
		return TRUE;
	}

	BOOL ret = FALSE;
	InstallModuleQueue::iterator it = m_installModuleQueue.begin();
	for( ; it != m_installModuleQueue.end(); it++ )
	{
		IInstallModule *module = (*it);
		if( _tcscmp( module->GetGuid(), guid ) != 0 )
		{
			continue;
		}

		module->SetPending(flag);
		ret = TRUE;
		break;
	}

	return ret;
}

BOOL InstallController::CancelModule( const TCHAR *guid )
{
	BOOL prepareRet = CancelPrepareModule( guid );
	BOOL installRet = CancelInstallModule( guid );

	NotifyPrepareSchedule();
	NotifyInstallSchedule();

	if( prepareRet || installRet )
		return TRUE;

	return FALSE;
}

FindPrepareModuleParameter InstallController::FindPrepareModule(const TCHAR * guid )
{
	FindPrepareModuleParameter findInfo;
	PrepareModuleQueue::iterator it = m_prepareModuleQueue.begin();
	for( ; it != m_prepareModuleQueue.end(); it++ )
	{
		IPrepareModule *module = *it;
		if( _tcsicmp( module->GetGuid(), guid ) == 0 )
		{
			break;
		}
	}

	findInfo.it = it;
	findInfo.guid = (TCHAR *)guid;
	return findInfo;
}

FindInstallModuleParameter InstallController::FindInstallModule(const TCHAR * guid )
{
	FindInstallModuleParameter findInfo;
	InstallModuleQueue::iterator it = m_installModuleQueue.begin();
	for( ; it != m_installModuleQueue.end(); it++ )
	{
		IInstallModule *module = *it;
		if( _tcsicmp( module->GetGuid(), guid ) == 0 )
		{
			break;
		}
	}

	findInfo.it = it;
	findInfo.guid = (TCHAR *)guid;
	return findInfo;
}

BOOL InstallController::CancelInstallModule( FindInstallModuleParameter findInfo )
{
	if( findInfo.it == m_installModuleQueue.end() )
		return FALSE;

	IInstallModule *module = *(findInfo.it);
	module->GetCallback()->OnFinished(CALLBACK_FINISHED_CANCEL, NULL);
	module->Cancel();
	delete module;
	m_installModuleQueue.erase(findInfo.it);

	if( m_installcontrollercallback )
		m_installcontrollercallback->OnInstallFinished(findInfo.guid, CALLBACK_FINISHED_CANCEL, NULL);

	return TRUE;
}

BOOL InstallController::CancelPrepareModule( FindPrepareModuleParameter findInfo )
{
	if( findInfo.it == m_prepareModuleQueue.end() )
		return FALSE;

	IPrepareModule *module = *(findInfo.it);
	module->GetCallback()->OnFinished(CALLBACK_FINISHED_CANCEL, NULL);
	module->Cancel();
	delete module;
	m_prepareModuleQueue.erase(findInfo.it);

	if( m_installcontrollercallback )
		m_installcontrollercallback->OnPrepareFinished(findInfo.guid, CALLBACK_FINISHED_CANCEL, NULL);

	return TRUE;
}

BOOL InstallController::CancelPrepareModule(const TCHAR * guid )
{
	if( m_currentPrepareModule && 
		_tcsicmp(m_currentPrepareModule->GetGuid(), guid) == 0 )
	{
		m_currentPrepareModule->Cancel();
		return TRUE;
	}

	FindPrepareModuleParameter findInfo = FindPrepareModule(guid);
	return CancelPrepareModule(findInfo);
}

BOOL InstallController::CancelInstallModule(const TCHAR * guid )
{
	if( m_currentInstallModule && 
		_tcsicmp(m_currentInstallModule->GetGuid(), guid) == 0 )
	{
		if( ModuleStatusINIT == m_currentInstallModule->GetModuleStatus() )
		{
			m_currentInstallModule->Cancel();
			m_currentInstallModule = NULL;
			return TRUE;
		}

		return FALSE;
	}

	FindInstallModuleParameter findInfo = FindInstallModule(guid);
	return CancelInstallModule(findInfo);
}

BOOL InstallController::Cancel( const TCHAR *guid)
{
	return CancelModule(guid);
}

IPrepareModule* InstallController::GetNextPrepareModule()
{
	if( m_prepareModuleQueue.empty() )
	{
		return NULL;
	}

	PrepareModuleQueue::iterator it = m_prepareModuleQueue.begin();
	IPrepareModule *module = (*it);
	m_prepareModuleQueue.pop_front();
	return module;
}

int gcount;
IPrepareModule* InstallController::GetCurrentPrepareModule()
{
	if( NULL == m_currentPrepareModule )
	{
		m_currentPrepareModule = GetNextPrepareModule();
	}
	else
	{
		if( m_currentPrepareModule->GetModuleStatus() == ModuleStatusFINISHED && m_currentPrepareModule->IsPending() == FALSE )
		{
			if( m_currentPrepareModule->IsCancel() || m_currentPrepareModule->IsError() )
			{
				delete m_currentPrepareModule;
				m_currentPrepareModule = NULL;

				MyPostMessage( m_scheduleModuleWnd->m_hWnd, WM_SCHEDULE_PREPAREMODULE, 0,0);
				return NULL;
			}

			BOOL bSend = FALSE;
			if( m_installModuleQueue.empty() )
				bSend = TRUE;

			IInstallModule *installModule = m_currentPrepareModule->GetInstallComponent();
			if( NULL != installModule )
			{
				installModule->SetPending(m_currentPrepareModule->IsPending());
				installModule->SetCallback(new MyInstallCallback(this->m_scheduleModuleWnd, installModule->GetGuid()) );
				m_installModuleQueue.push_back(installModule);
			}

			delete m_currentPrepareModule;
			m_currentPrepareModule = GetNextPrepareModule();
			if( bSend )
			{
				MyPostMessage( m_scheduleModuleWnd->m_hWnd, WM_SCHEDULE_INSTALLMODULE, 0,0);
			}
		}
	}

	return m_currentPrepareModule;
}

DWORD WINAPI InstallFunction(LPVOID *para)
{
	IInstallModule *installModule = reinterpret_cast<IInstallModule *>(para);
	ATLTRACE(_T("InstallProcess\n"));
	installModule->Install();
	return 1;
}

DWORD WINAPI PrepareFunction(LPVOID *para)
{
	IPrepareModule *prepareModule = reinterpret_cast<IPrepareModule *>(para);
	ATLTRACE(_T("PrepareProcess\n"));
	prepareModule->Prepare();
	return 1;
}

BOOL InstallController::ProcessPrepare()
{
	ATLTRACE(_T("ProcessPrepare\n") );
	IPrepareModule *currentPrepareModule = GetCurrentPrepareModule();
	if( NULL == currentPrepareModule )
	{
		ATLTRACE(_T("currentProcessPrepare is FALSE\n") );
		return FALSE;
	}

	if (ModuleStatusINIT == currentPrepareModule->GetModuleStatus() && currentPrepareModule->IsPending() == FALSE)
	{
		ATLTRACE(_T("Prepare Insert Queue\n") );
		currentPrepareModule->SetModuleStatus(ModuleStatusPROCESS);
		currentPrepareModule->GetCallback()->OnStart();

		if( ERROR_SUCCESS != NPQueueWorkItem( m_prepareThreadPool, (LPTHREAD_START_ROUTINE) PrepareFunction, (LPVOID)currentPrepareModule ) )
		{
			ATLTRACE(_T("NPQueueWorkItem Error in Prepare\n"));
			return FALSE;
		}
	}
	return TRUE;
}

IInstallModule* InstallController::GetNextInstallModule()
{
	if( m_installModuleQueue.empty() )
	{
		ATLTRACE(_T("Install: Module Queue is Empty\n"));
		return NULL;
	}

	InstallModuleQueue::iterator it = m_installModuleQueue.begin();
	IInstallModule *module = (*it);
	m_installModuleQueue.pop_front();
	return module;
}

IInstallModule* InstallController::GetCurrentInstallModule()
{
	if( NULL == m_currentInstallModule )
	{
		m_currentInstallModule = GetNextInstallModule();
	}
	else
	{
		if( m_currentInstallModule->GetModuleStatus() == ModuleStatusFINISHED && m_currentInstallModule->IsPending() == FALSE )
		{
			ATLTRACE(_T("Install: GetNextModuleQueue\n"));
			IInstallModule *module = m_currentInstallModule;
			m_currentInstallModule = GetNextInstallModule();
			delete module;
		}
	}

	return m_currentInstallModule;
}

BOOL InstallController::ProcessInstall()
{
	IInstallModule *currentInstallModule = GetCurrentInstallModule();
	if( NULL == currentInstallModule )
	{	
		ATLTRACE(_T("Install: Module None\n"));
		return FALSE;
	}

	ATLTRACE(_T("Install: Module Status %d\n"), currentInstallModule->GetModuleStatus());
	if( currentInstallModule->GetModuleStatus() == ModuleStatusINIT && currentInstallModule->IsPending() == FALSE )
	{
		ATLTRACE(_T("Install: Module Insert Queue %d\n"), currentInstallModule->GetModuleStatus());
		currentInstallModule->SetModuleStatus(ModuleStatusPROCESS);

		CALLBACK_FINISHED_STATUS status = CALLBACK_FINISHED_SUCCESS;
		CALLBACK_RETURN ret = m_installcontrollercallback->OnInstallStart( currentInstallModule->GetGuid() );

		switch( ret )
		{
		case CALLBACK_RETURN_CONTINUE:
			NPQueueWorkItem( m_installThreadPool, (LPTHREAD_START_ROUTINE) InstallFunction, (LPVOID)currentInstallModule );
			break;

		case CALLBACK_RETURN_CANCEL:
			status = CALLBACK_FINISHED_CANCEL;
			break;

		case CALLBACK_RETURN_ERROR:
			status = CALLBACK_FINISHED_ERROR;
			break;
			
		}

		if( CALLBACK_FINISHED_SUCCESS != status )
		{
			currentInstallModule->SetModuleStatus(ModuleStatusFINISHED);
			CALLBACK_RETURN callback_ret = m_installcontrollercallback->OnInstallFinished( currentInstallModule->GetGuid(), status, NULL );
			//  ġ Ͽ üũѴ.
			if( callback_ret != CALLBACK_RETURN_ALLDONE )
			{
				MyPostMessage( m_scheduleModuleWnd->m_hWnd, WM_SCHEDULE_INSTALLMODULE, 0, 0);
			}
		}
	}

	return TRUE;
}

CALLBACK_RETURN InstallController::OnProgress( const TCHAR *guid,
						   CALLBACK_PROGRESS_SIZE currnetSize,
						   CALLBACK_PROGRESS_SIZE maxSize,
						   CALLBACK_PROGRESS_STATUS status )
{
	return m_installcontrollercallback->OnProgress(guid, currnetSize, maxSize, status);
}

CALLBACK_RETURN InstallController::OnPrepareFinished( 
									const TCHAR *guid, CALLBACK_FINISHED_STATUS status, ERRORINFO *info )
{
	return m_installcontrollercallback->OnPrepareFinished(guid, status, info);
}

CALLBACK_RETURN InstallController::OnFinished( 
									const TCHAR *guid, CALLBACK_FINISHED_STATUS status, ERRORINFO *info )
{
	return m_installcontrollercallback->OnInstallFinished(guid, status, info);
}

CALLBACK_RETURN InstallController::OnPrepareStart( const TCHAR *guid )
{
	return m_installcontrollercallback->OnPrepareStart(guid);
}

CALLBACK_RETURN InstallController::OnInstallStart( const TCHAR *guid )
{
	return m_installcontrollercallback->OnInstallStart(guid);
}